/*global
  XLAB:false
  global:false
  console:false
  clone:false
  Shader:false
  ShaderProgram:false
*/

var PLINE_NONE = 0;
var PLINE_FULL = 1;
var PLINE_PART = 2;

var COLORMODE_NONE = 0;
var COLORMODE_TEXIMAGE = 1;
var COLORMODE_CUBEMAP = 2;
var COLORMODE_LOCALCOLOR = 3;

var TEXIMAGE_TEXTURE_UNIT = 0;
var CUBEMAP_TEXTURE_UNIT = 0;

var PATTERN_NONE = 0;
var PATTERN_SQUARE = 1;
var PATTERN_CIRCLE = 2;
var PATTERN_SUPERELLIPSE = 3;
var PATTERN_STRIPE0 = 4;
var PATTERN_STRIPE1 = 5;


var def = [
    // 0
    [
        '',
        '#define STEREOGRAPHIC_PROJECTION\n'
        ],
    // 1
    [
        '',
        '#define SPECULAR\n'
        ],
    // 2
    [
        '',
        '#define CLIP\n'
        ],
    // 3
    [
        '',
        '#define PLINE_FULL\n',
        '#define PLINE_PART\n'
        ],
    // 4
    [
        '',
        '#define TEXIMAGE\n',
        '#define CUBEMAP\n',
        '#define LOCALCOLOR\n'
        ],
    // 5
    [
        '',
        '#define PATTERN_SQUARE\n',
        '#define PATTERN_CIRCLE\n',
        '#define PATTERN_SUPERELLIPSE\n',
        '#define PATTERN_STRIPE0\n',
        '#define PATTERN_STRIPE1\n'
        ],
    // 6
    [
        '',
        '#define CUTAWAY\n'
        ],
    // 7
    [
        '',
        '#define TEXCLIP_Y\n',
        '#define TEXCLIP_X\n',
        '#define TEXCLIP_X\n#define TEXCLIP_Y\n'
        ],
    // 8
    [
        '',
        '#define have_OES_standard_derivatives\n'
        ]
    ];

//------------------------------------------------------------------------------------

/*
  XShader constructor
*/
function XShader(gl, mode, parameters, src, urls) {
  "use strict";

  this.gl = gl;
  this.src = src;
  this.urls = urls;
  this.parameters = parameters;
  this._init(mode);
}



XShader.prototype._init = function( mode ) {
  "use strict";

  if (mode === undefined) {
    throw "mode is undefined";
  }

  if (this.shaderProgram && mode == this.mode) { return; }

  var gl = this.gl;

  this.mode = clone( mode );

  var a0 = this.mode.stereographic ? 1 : 0;
  var a1 = this.mode.specular ? 1 : 0;
  var a2 = this.mode.clip ? 1 : 0;
  var a3 = this.mode.pline;
  var a4 = this.mode.color;
  var a5 = this.mode.pattern;
  var a6 = this.mode.cutaway ? 1 : 0;
  var a7 = this.mode.texclip;
  var a8 = this.mode.have_OES_standard_derivatives ? 1 : 0;

  var define = [
      def[0][a0], def[1][a1], def[2][a2], def[3][a3], def[4][a4], def[5][a5], def[6][a6],
      def[7][a7],
      def[8][a8]
      ];

  var vertex_shader_src = clone(define);
  vertex_shader_src.push(this.src[0]);

  var fragment_shader_src = clone(define);
  fragment_shader_src.push(this.src[1]);

  this.shaderProgram = XLAB.createShader( gl, [ vertex_shader_src, fragment_shader_src ], this.urls );

  this.shaderProgram.use(gl);
  this._initAttributes();
  this._initUniforms();
  this.sync();
};
  
XShader.prototype._initAttributes = function() {
  "use strict";

  var gl = this.gl;
  this._attribute = {};
  var S = this.shaderProgram;

  // get attribute ids and enable vertex arrays

  // position
  this._attribute.position = S.getAttribLocation(gl, "position");
  if (this._attribute.position == -1) {
    throw "missing attribute: position";
  }

  // normal
  this._attribute.normal = S.getAttribLocation(gl, "normal");
  if (this._attribute.normal == -1) {
    throw "missing attribute: normal";
  }

  // texture
  if ( this.mode.pattern != PATTERN_NONE ||
       this.mode.pline !== 0 ||
       this.mode.color == COLORMODE_TEXIMAGE ||
       this.mode.texclip !== 0) {
    this._attribute.texture = S.getAttribLocation(gl, "texture");
    if (this._attribute.texture == -1) {
      throw "missing attribute: texture";
    }
  }
  if (this.mode.color == COLORMODE_LOCALCOLOR) {
    // color
    this._attribute.color = S.getAttribLocation(gl, "color");
    if (this._attribute.color == -1) {
      throw "missing attribute: color";
    }
  }
};

XShader.prototype._initUniforms = function() {
  "use strict";

  var gl = this.gl;
  var uniformTransform = [
      "projective_transform",
      "modelview_transform",
      "texture_transform0",
      "texture_transform1"
      ];

  this._uniform = {};
  var S = this.shaderProgram;

  // get uniform ids for transforms
  for (var i in uniformTransform) {
    var keyx = uniformTransform[i];
    this._uniform[keyx] = S.getUniformLocation(gl, keyx);
  }

  for (var key in this.parameters) {
    this._uniform[key] = S.getUniformLocation(gl, key);
  }

  switch (this.mode.color)
  {
    case COLORMODE_TEXIMAGE:
    this._uniform.teximage = S.getUniformLocation(gl, 'teximage');
    break;
    case COLORMODE_CUBEMAP:
    this._uniform.cubemap = S.getUniformLocation(gl, 'cubemap');
    break;
  }
};


XShader.prototype.use = function() {
  "use strict";
  if (this.shaderProgram === undefined) {
    throw "shaderProgram is undefined";
  }
  this.shaderProgram.use(this.gl);
};

XShader.prototype.attribute = function() {
  "use strict";
  return this._attribute;
};

XShader.prototype.destroy = function() {
  "use strict";
  // delete old shader
  this.gl.useProgram(null);
  this.shaderProgram.destroy(this.gl);
  this.shaderProgram = undefined;
  this._uniform = undefined;
  this._attribute = undefined;
};

XShader.prototype.sync = function() {
  "use strict";

  var gl = this.gl;
  var S = this.shaderProgram;
  S.use(gl);

  var parameters = this.parameters;

  gl.uniform3fv( this._uniform.camera_position, parameters.camera_position );
  gl.uniform3fv( this._uniform.light_direction, parameters.light_direction );
  gl.uniform3fv( this._uniform.ambient, this.parameters.ambient );
  gl.uniform3fv( this._uniform.diffuse_front, this.parameters.diffuse_front );
  gl.uniform3fv( this._uniform.diffuse_back, this.parameters.diffuse_back );

  if (this.mode.specular)
  {
    gl.uniform1f(  this._uniform.shininess, parameters.shininess);
    gl.uniform3fv( this._uniform.specular, parameters.specular );
  }

  if (this.mode.pattern !== PATTERN_NONE)
  {
    gl.uniform1f(  this._uniform.pattern_width, parameters.pattern_width);
    gl.uniform2fv(  this._uniform.pattern_exponent, parameters.pattern_exponent);
    gl.uniform3fv( this._uniform.pattern_front_color, this.parameters.pattern_front_color);
    gl.uniform3fv( this._uniform.pattern_back_color, this.parameters.pattern_back_color);
  }

  /*
  if (this.mode.pattern == PATTERN_SQUARE )
  {
    gl.uniform1f(  this._uniform.lattice_width, parameters.lattice_width);
    //gl.uniform2fv( this._uniform.lattice_frequency, parameters.lattice_frequency);
  }

  if (this.mode.pattern == PATTERN_STRIPE0 || this.mode.pattern == PATTERN_STRIPE1 ) {
    gl.uniform1f(  this._uniform.stripe_width, parameters.stripe_width);
    //gl.uniform2fv( this._uniform.stripe_frequency, parameters.stripe_frequency);
  }
*/

  if (this.mode.clip)
  {
    gl.uniform4fv( this._uniform.clipplane, parameters.clipplane);
  }

  if (this.mode.pline !== 0)
  {
    gl.uniform2fv( this._uniform.pline_width, parameters.pline_width );
    gl.uniform2fv( this._uniform.pline_position, parameters.pline_position );
    gl.uniform3fv( this._uniform.pline_x_color, this.parameters.pline_x_color );
    gl.uniform3fv( this._uniform.pline_y_color, this.parameters.pline_y_color );
  }

  switch (this.mode.color)
  {
    case COLORMODE_TEXIMAGE:
    gl.uniform1i( this._uniform.teximage, TEXIMAGE_TEXTURE_UNIT );
    break;
    case COLORMODE_CUBEMAP:
    gl.uniform1i( this._uniform.cubemap, CUBEMAP_TEXTURE_UNIT );
    break;
    case COLORMODE_LOCALCOLOR:
    gl.uniform3fv( this._uniform.localcolor0, parameters.localcolor0 );
    gl.uniform3fv( this._uniform.localcolor1, parameters.localcolor1 );
    break;
  }

};

XShader.prototype.setProjectiveTransform = function(matrix) {
  "use strict";
  this.use();
  this.gl.uniformMatrix4fv(this._uniform.projective_transform, false, matrix);
};

XShader.prototype.setModelviewTransform = function(matrix) {
  "use strict";
  this.use();
  this.gl.uniformMatrix4fv(this._uniform.modelview_transform, false, matrix);
};

XShader.prototype.set_texture_transform0 = function(matrix) {
  "use strict";
  this.use();
  this.gl.uniformMatrix3fv(this._uniform.texture_transform0, false, matrix);
};

XShader.prototype.set_texture_transform1 = function(matrix) {
  "use strict";
  this.use();
  this.gl.uniformMatrix3fv(this._uniform.texture_transform1, false, matrix);
};
